mod config;
mod meta;

pub use config::ManagerConfig;
pub use meta::MetaManager;

use super::{Backup, State};
use crate::other;
use chrono::SubsecRound;
use chrono::Utc;
use serde::Deserialize;
use serde::Serialize;
use std::fs::File;
use std::io;
use std::path::PathBuf;

/// Manages a single backup layer consisting of one or more chains of backups.
pub struct Manager<T>
where
    T: Clone + Serialize + for<'de> Deserialize<'de> + std::fmt::Debug,
{
    backup_dir: PathBuf,
    dirs: Vec<(PathBuf, PathBuf)>,
    default_metadata: T,
    chain_len: u64,
    chains_to_keep: u64,
    frequency: chrono::Duration,
    chains: Vec<Vec<Backup<T>>>,
}

impl<T> Manager<T>
where
    T: Clone + Serialize + for<'de> Deserialize<'de> + std::fmt::Debug,
{
    const METADATA_FILE: &str = "alex.json";

    pub fn new<P: Into<PathBuf>>(
        backup_dir: P,
        dirs: Vec<(PathBuf, PathBuf)>,
        metadata: T,
        chain_len: u64,
        chains_to_keep: u64,
        frequency: chrono::Duration,
    ) -> Self {
        Self {
            backup_dir: backup_dir.into(),
            dirs,
            default_metadata: metadata,
            chain_len,
            chains_to_keep,
            frequency,
            chains: Vec::new(),
        }
    }

    /// Create a new backup, either full or incremental, depending on the state of the current
    /// chain.
    pub fn create_backup(&mut self) -> io::Result<()> {
        // We start a new chain if the current chain is complete, or if there isn't a first chain
        // yet
        if let Some(current_chain) = self.chains.last() {
            let current_chain_len: u64 = current_chain.len().try_into().unwrap();

            if current_chain_len >= self.chain_len {
                self.chains.push(Vec::new());
            }
        } else {
            self.chains.push(Vec::new());
        }

        let current_chain = self.chains.last_mut().unwrap();

        let mut backup = if !current_chain.is_empty() {
            let previous_backup = current_chain.last().unwrap();
            let previous_state = State::from(current_chain.iter().map(|b| &b.delta));

            Backup::create_from(
                previous_state,
                previous_backup.start_time,
                &self.backup_dir,
                &self.dirs,
            )?
        } else {
            Backup::create(&self.backup_dir, &self.dirs)?
        };

        backup.set_metadata(self.default_metadata.clone());

        current_chain.push(backup);

        self.save()?;

        Ok(())
    }

    /// Delete all backups associated with outdated chains, and forget those chains.
    pub fn remove_old_backups(&mut self) -> io::Result<()> {
        let chains_to_store: usize = self.chains_to_keep.try_into().unwrap();

        if chains_to_store < self.chains.len() {
            let mut remove_count: usize = self.chains.len() - chains_to_store;

            // We only count finished chains towards the list of stored chains
            let chain_len: usize = self.chain_len.try_into().unwrap();
            if self.chains.last().unwrap().len() < chain_len {
                remove_count -= 1;
            }

            for chain in self.chains.drain(..remove_count) {
                for backup in chain {
                    let path = Backup::path(&self.backup_dir, backup.start_time);
                    std::fs::remove_file(path)?;
                }
            }

            self.save()?;
        }

        Ok(())
    }

    /// Write the in-memory state to disk.
    pub fn save(&self) -> io::Result<()> {
        let json_file = File::create(self.backup_dir.join(Self::METADATA_FILE))?;
        serde_json::to_writer(json_file, &self.chains)?;

        Ok(())
    }

    /// Overwrite the in-memory state with the on-disk state.
    pub fn load(&mut self) -> io::Result<()> {
        let json_file = match File::open(self.backup_dir.join(Self::METADATA_FILE)) {
            Ok(f) => f,
            Err(e) => {
                // Don't error out if the file isn't there, it will be created when necessary
                if e.kind() == io::ErrorKind::NotFound {
                    self.chains = Vec::new();

                    return Ok(());
                } else {
                    return Err(e);
                }
            }
        };

        self.chains = serde_json::from_reader(json_file)?;

        Ok(())
    }

    /// Calculate the next time a backup should be created. If no backup has been created yet, it
    /// will return now.
    pub fn next_scheduled_time(&self) -> chrono::DateTime<Utc> {
        if let Some(last_chain) = self.chains.last() {
            if let Some(last_backup) = last_chain.last() {
                return last_backup.start_time + self.frequency;
            }
        }

        chrono::offset::Utc::now()
    }

    /// Restore the backup with the given start time by restoring its chain up to and including the
    /// backup, in order.
    pub fn restore_backup(
        &self,
        start_time: chrono::DateTime<Utc>,
        dirs: &Vec<(PathBuf, PathBuf)>,
    ) -> io::Result<()> {
        // Iterate over each chain, skipping elements until the element with the given start time
        // is possibly found.
        for chain in &self.chains {
            // If we find the element in the chain, restore the entire chain up to and including
            // the element
            if let Some(index) = chain
                .iter()
                .position(|b| b.start_time.trunc_subsecs(0) == start_time)
            {
                for backup in chain.iter().take(index + 1) {
                    backup.restore(&self.backup_dir, dirs)?;
                }

                return Ok(());
            }
        }

        Err(other("Unknown backup."))
    }

    /// Get a reference to the underlying chains
    pub fn chains(&self) -> &Vec<Vec<Backup<T>>> {
        &self.chains
    }
}
