mod backup;
mod run;

pub use crate::backup::MetaManager;
pub use crate::server::Metadata;
pub use backup::{BackupArgs, BackupCommands};
pub use run::RunArgs;

use crate::backup::ManagerConfig;
use crate::server::ServerType;
use clap::{Parser, Subcommand};
use std::io;
use std::path::PathBuf;

#[derive(Subcommand)]
pub enum Commands {
    /// Run the server
    Run(RunArgs),
    /// Interact with the backup system without starting a server
    Backup(BackupArgs),
}

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    pub command: Commands,
    /// Directory where configs are stored, and where the server will run
    #[arg(
        long,
        value_name = "CONFIG_DIR",
        default_value = ".",
        env = "ALEX_CONFIG_DIR",
        global = true
    )]
    pub config: PathBuf,
    /// Directory where world files will be saved
    #[arg(
        long,
        value_name = "WORLD_DIR",
        default_value = "../worlds",
        env = "ALEX_WORLD_DIR",
        global = true
    )]
    pub world: PathBuf,
    /// Directory where backups will be stored
    #[arg(
        long,
        value_name = "BACKUP_DIR",
        default_value = "../backups",
        env = "ALEX_BACKUP_DIR",
        global = true
    )]
    pub backup: PathBuf,

    /// What backup layers to employ, provided as a list of tuples name,frequency,chains,chain_len
    /// delimited by semicolons (;).
    #[arg(long, env = "ALEX_LAYERS", global = true, value_delimiter = ';')]
    pub layers: Vec<ManagerConfig>,

    /// Type of server
    #[arg(long, default_value = "unknown", env = "ALEX_SERVER", global = true)]
    pub server: ServerType,
    /// Version string for the server, e.g. 1.19.4-545
    #[arg(long, default_value = "", env = "ALEX_SERVER_VERSION", global = true)]
    pub server_version: String,
}

impl Cli {
    pub fn run(&self) -> io::Result<()> {
        match &self.command {
            Commands::Run(args) => args.run(self),
            Commands::Backup(args) => args.run(self),
        }
    }

    /// Convenience method to initialize backup manager from the cli arguments
    pub fn meta(&self) -> io::Result<MetaManager<Metadata>> {
        let metadata = Metadata {
            server_type: self.server,
            server_version: self.server_version.clone(),
        };
        let dirs = vec![
            (PathBuf::from("config"), self.config.canonicalize()?),
            (PathBuf::from("worlds"), self.world.canonicalize()?),
        ];
        let mut meta = MetaManager::new(self.backup.canonicalize()?, dirs, metadata);
        meta.add_all(&self.layers)?;

        Ok(meta)
    }
}
