use crate::backup::MetaManager;
use crate::server::Metadata;
use std::io::Write;
use std::process::Child;

pub struct ServerProcess {
    child: Child,
    pub backups: MetaManager<Metadata>,
}

impl ServerProcess {
    pub fn new(manager: MetaManager<Metadata>, child: Child) -> ServerProcess {
        ServerProcess {
            child,
            backups: manager,
        }
    }

    pub fn send_command(&mut self, cmd: &str) -> std::io::Result<()> {
        match cmd.trim() {
            "stop" | "exit" => self.stop()?,
            "backup" => self.backup()?,
            s => self.custom(s)?,
        }

        Ok(())
    }

    fn custom(&mut self, cmd: &str) -> std::io::Result<()> {
        let mut stdin = self.child.stdin.as_ref().unwrap();
        stdin.write_all(format!("{}\n", cmd.trim()).as_bytes())?;
        stdin.flush()?;

        Ok(())
    }

    pub fn stop(&mut self) -> std::io::Result<()> {
        self.custom("stop")?;
        self.child.wait()?;

        Ok(())
    }

    pub fn kill(&mut self) -> std::io::Result<()> {
        self.child.kill()
    }

    /// Perform a backup by disabling the server's save feature and flushing its data, before
    /// creating an archive file.
    pub fn backup(&mut self) -> std::io::Result<()> {
        let layer_name = String::from(self.backups.next_scheduled_layer().unwrap());
        self.custom(&format!("say starting backup for layer '{}'", layer_name))?;

        // Make sure the server isn't modifying the files during the backup
        self.custom("save-off")?;
        self.custom("save-all")?;

        // TODO implement a better mechanism
        // We wait some time to (hopefully) ensure the save-all call has completed
        std::thread::sleep(std::time::Duration::from_secs(10));

        let start_time = chrono::offset::Utc::now();
        let res = self.backups.perform_backup_cycle();

        // The server's save feature needs to be enabled again even if the archive failed to create
        self.custom("save-on")?;
        self.custom("save-all")?;

        let duration = chrono::offset::Utc::now() - start_time;
        let duration_str = format!(
            "{}m{}s",
            duration.num_seconds() / 60,
            duration.num_seconds() % 60
        );

        if res.is_ok() {
            self.custom(&format!(
                "say backup created for layer '{}' in {}",
                layer_name, duration_str
            ))?;
        } else {
            self.custom(&format!(
                "an error occured after {} while creating backup for layer '{}'",
                duration_str, layer_name
            ))?;
        }

        res
    }
}
